#include <assert.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdbool.h>
#include <stdint.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <gpio_sys.h>

#define IN	0
#define OUT	1

#define BUF_SIZE	3
#define DIRECTION_MAX	35
#define VALUE_MAX	30

static int gpio_export(int pin);
static int gpio_unexport(int pin);
static int gpio_set_direction(int pin, int dir);

int gpio_open(const struct gpio_sys *gpio)
{
    int res;

    assert(NULL != gpio);

    syslog(LOG_DEBUG, "Start opening gpio pin %d.\n", gpio->pin);
    res = gpio_export(gpio->pin);
    if(res < 0) {
        syslog(LOG_ERR, "Unable to export pin %d.\n", gpio->pin);
        return res;
    }

    usleep(100000); /* Wait until sysfs is exported. */

    syslog(LOG_DEBUG, "Gpio pin %d successfully opened.\n", gpio->pin);
    return 0;
}

int gpio_close(const struct gpio_sys *gpio)
{
    int res;

    assert(NULL != gpio);

    syslog(LOG_DEBUG, "Start closing gpio pin %d\n.", gpio->pin);
    res = gpio_unexport(gpio->pin);
    if(res < 0) {
        syslog(LOG_ERR, "Unable to unexport pin %d.\n", gpio->pin);
        return res;
    }

    syslog(LOG_DEBUG, "Gpio pin %d successfully closed.\n", gpio->pin);
    return 0;
}

int gpio_direction(struct gpio_sys *gpio, int direction)
{
    int res;

    assert(NULL != gpio);
    syslog(LOG_DEBUG, "Start setting gpio direction for pin %d\n.", gpio->pin);

    res = gpio_set_direction(gpio->pin, direction);
    if(res < 0) {
        syslog(LOG_ERR, "Unable to set direction for pin %d.\n", gpio->pin);
        return res;
    }
    gpio->direction = direction;

    syslog(LOG_DEBUG, "Successfully set gpio direction for pin %d\n.", gpio->pin);
    return 0;
}

int gpio_read(const struct gpio_sys *gpio)
{
    char path[VALUE_MAX];
    char value_str[BUF_SIZE];
    int fd;

    assert(NULL != gpio);
    snprintf(path, VALUE_MAX, "/sys/class/gpio/gpio%d/value", gpio->pin);
    fd = open(path, O_RDONLY);
    if(-1 == fd) {
        syslog(LOG_ERR, "Failed to open pin %d value for reading.\n", gpio->pin);
        return -1;
    }

    if(-1 == read(fd, value_str, BUF_SIZE)) {
        syslog(LOG_ERR, "Failed to read value for pin %d.\n", gpio->pin);
        return -1;
    }
    close(fd);

    return(atoi(value_str));
}

int gpio_write(const struct gpio_sys *gpio, int value)
{
    static const char s_values_str[] = "01";
    char path[VALUE_MAX];
    int fd;

    assert(NULL != gpio);

    snprintf(path, VALUE_MAX, "/sys/class/gpio/gpio%d/value", gpio->pin);
    fd = open(path, O_WRONLY);
    if(-1 == fd) {
        syslog(LOG_ERR, "Failed to open pin %d value for writing.\n", gpio->pin);
        return -1;
    }

    if(1 != write(fd, &s_values_str[LOW == value ? 0 : 1], 1)) {
        syslog(LOG_ERR, "Failed to write value for pin %d.\n", gpio->pin);
        return -1;
    }
    close(fd);

    return 0;
}

static int gpio_export(int pin)
{
    char buffer[BUF_SIZE];
    ssize_t bytes_written;
    int fd;

    fd = open("/sys/class/gpio/export", O_WRONLY);
    if(-1 == fd) {
        syslog(LOG_ERR, "Failed to open export for writing.\n");
        return -1;
    }

    bytes_written = snprintf(buffer, BUF_SIZE, "%d", pin);
    write(fd, buffer, bytes_written);
    close(fd);

    return 0;
}

static int gpio_unexport(int pin)
{
    char buffer[BUF_SIZE];
    ssize_t bytes_written;
    int fd;

    fd = open("/sys/class/gpio/unexport", O_WRONLY);
    if(-1 == fd) {
        syslog(LOG_ERR, "Failed to open unexport for writing.\n");
        return -1;
    }

    bytes_written = snprintf(buffer, BUF_SIZE, "%d", pin);
    write(fd, buffer, bytes_written);
    close(fd);

    return 0;
}

static int gpio_set_direction(int pin, int dir)
{
    static const char directions_str[]  = "in\0out";
    char path[DIRECTION_MAX];
    int fd;

    snprintf(path, DIRECTION_MAX, "/sys/class/gpio/gpio%d/direction", pin);
    fd = open(path, O_WRONLY);
    if(-1 == fd) {
        syslog(LOG_ERR, "Failed to open direction for writing.\n");
        return -1;
    }

    if(-1 == write(fd, &directions_str[IN == dir ? 0 : 3], IN == dir ? 2 : 3)) {
        syslog(LOG_ERR, "Failed to set direction.\n");
        return -1;
    }
    close(fd);

    return 0;
}
